#!/usr/bin/env python3

from fabric import Connection, Config
from invoke.exceptions import UnexpectedExit
from pathlib import Path
from sys import stderr
from io import StringIO
from json import loads
from argparse import ArgumentParser
from dataclasses import dataclass
from dataclass_wizard import fromdict
from ipaddress import IPv4Interface, IPv4Network, IPv6Interface, IPv6Network, AddressValueError, NetmaskValueError
from itertools import combinations
from wireguard_tools import WireguardConfig, WireguardPeer, WireguardKey

@dataclass
class WovenArgs:
    config: str

@dataclass
class WovenNode:
    listen_address: str
    listen_gateway: str
    interface_name: str
    routed_ipv4_subnets: list[str]
    routed_ipv6_subnets: list[str]

@dataclass
class WovenConfig:
    ptp_ipv4_network: str
    ptp_ipv6_network: str
    min_port: int
    max_port: int
    nodes: dict[str, WovenNode]
    ptp_ipv4_subnet: int = 30
    ptp_ipv6_subnet: int = 64

def generate_wg_configs(config: WovenConfig):
    try:
        ptp_ipv4_network = IPv4Network(config.ptp_ipv4_network)
    except AddressValueError:
        raise ValueError("invalid IPv4 PtP network address")
    except NetmaskValueError:
        raise ValueError("invalid IPv4 PtP network subnet")
    try:
        ptp_ipv6_network = IPv6Network(config.ptp_ipv6_network)
    except AddressValueError:
        raise ValueError("invalid IPv6 PtP network address")
    except NetmaskValueError:
        raise ValueError("invalid IPv6 PtP network subnet")
    
    cs = { id: Connection(node.listen_address, user = "root", config = Config(overrides = { "run": { "hide": True } })) for id, node in config.nodes.items() }
    for c in cs.values():
        c.run("for f in /etc/wireguard/*-loop.conf; do systemctl stop wg-quick@$(basename $f .conf).service; done")
        try:
            c.run("rm /etc/wireguard/*-loop.conf")
        except UnexpectedExit:
            pass
    
    ptp_ipv4_network_iter = ptp_ipv4_network.subnets(new_prefix = config.ptp_ipv4_subnet)
    ptp_ipv6_network_iter = ptp_ipv6_network.subnets(new_prefix = config.ptp_ipv6_subnet)
    port_iter = iter(range(config.min_port, config.max_port))
    
    for (id_a, node_a), (id_b, node_b) in combinations(config.nodes.items(), 2):
        try:
            ptp_ipv4_network = next(ptp_ipv4_network_iter)
        except StopIteration:
            raise ValueError("not enough IPv4 PtP networks to assign")
        try:
            ptp_ipv6_network = next(ptp_ipv6_network_iter)
        except StopIteration:
            raise ValueError("not enough IPv6 PtP networks to assign")
        try:
            port = next(port_iter)
        except StopIteration:
            raise ValueError("not enough ports to assign")
        
        ipv4_iter = ptp_ipv4_network.hosts()
        try:
            ipv4_a = next(ipv4_iter)
            ipv4_b = next(ipv4_iter)
        except StopIteration:
            raise ValueError("not enough IPv4 addresses in each PtP network")
        
        ipv6_iter = ptp_ipv6_network.hosts()
        try:
            ipv6_a = next(ipv6_iter)
            ipv6_b = next(ipv6_iter)
        except StopIteration:
            raise ValueError("not enough IPv6 addresses in each PtP network")
        
        key_a = WireguardKey.generate()
        key_a_pub = key_a.public_key()
        
        key_b = WireguardKey.generate()
        key_b_pub = key_b.public_key()
        
        name_a = f"{id_a}-{id_b}-loop"
        name_b = f"{id_b}-{id_a}-loop"
        
        preup_a = [f"ip ro replace {node_b.listen_address}/32 dev {node_a.interface_name} via {node_a.listen_gateway} metric 10 src {node_a.listen_address}"]
        predown_a = [f"ip ro del {node_b.listen_address}/32 dev {node_a.interface_name} via {node_a.listen_gateway} metric 10 src {node_a.listen_address}"]
        postup_a = [f"ip ro replace {sn} dev {name_a} via {ipv4_b} metric 10" for sn in node_b.routed_ipv4_subnets] + [f"ip -6 ro replace {sn} dev {name_a} via {ipv6_b} metric 10" for sn in node_b.routed_ipv6_subnets]
        postdown_a = [f"ip ro del {sn} dev {name_a} via {ipv4_b} metric 10" for sn in node_b.routed_ipv4_subnets] + [f"ip -6 ro del {sn} dev {name_a} via {ipv6_b} metric 10" for sn in node_b.routed_ipv6_subnets]
        
        config_a = WireguardConfig(
            addresses = [IPv4Interface(f"{ipv4_a}/{ptp_ipv4_network.prefixlen}"), IPv6Interface(f"{ipv6_a}/{ptp_ipv6_network.prefixlen}")],
            listen_port = port,
            private_key = key_a,
            table = False,
            preup = preup_a,
            predown = predown_a,
            postup = postup_a,
            postdown = postdown_a,
            peers = {
                key_b_pub: WireguardPeer(
                    public_key = key_b_pub,
                    allowed_ips = ["0.0.0.0/0", "::/0"],
                    endpoint_host = node_b.listen_address,
                    endpoint_port = port,
                    persistent_keepalive = 20
                )
            }
        )
        
        preup_b = [f"ip ro replace {node_a.listen_address}/32 dev {node_b.interface_name} via {node_b.listen_gateway} metric 10 src {node_b.listen_address}"]
        predown_b = [f"ip ro del {node_a.listen_address}/32 dev {node_b.interface_name} via {node_b.listen_gateway} metric 10 src {node_b.listen_address}"]
        postup_b = [f"ip ro replace {sn} dev {name_b} via {ipv4_a} metric 10" for sn in node_a.routed_ipv4_subnets] + [f"ip -6 ro replace {sn} dev {name_b} via {ipv6_a} metric 10" for sn in node_a.routed_ipv6_subnets]
        postdown_b = [f"ip ro del {sn} dev {name_b} via {ipv4_a} metric 10" for sn in node_a.routed_ipv4_subnets] + [f"ip -6 ro del {sn} dev {name_b} via {ipv6_a} metric 10" for sn in node_a.routed_ipv6_subnets]
        
        config_b = WireguardConfig(
            addresses = [IPv4Interface(f"{ipv4_b}/{ptp_ipv4_network.prefixlen}"), IPv6Interface(f"{ipv6_b}/{ptp_ipv6_network.prefixlen}")],
            listen_port = port,
            private_key = key_b,
            table = False,
            preup = preup_b,
            predown = predown_b,
            postup = postup_b,
            postdown = postdown_b,
            peers = {
                key_a_pub: WireguardPeer(
                    public_key = key_a_pub,
                    allowed_ips = ["0.0.0.0/0", "::/0"],
                    endpoint_host = node_a.listen_address,
                    endpoint_port = port,
                    persistent_keepalive = 20
                )
            }
        )
        
        
        cs[id_a].put(StringIO(config_a.to_wgconfig(wgquick_format = True)), f"/etc/wireguard/{name_a}.conf")
        cs[id_a].run(f"systemctl start wg-quick@{name_a}.service")
        
        cs[id_b].put(StringIO(config_b.to_wgconfig(wgquick_format = True)), f"/etc/wireguard/{name_b}.conf")
        cs[id_b].run(f"systemctl start wg-quick@{name_b}.service")


def main():
    try:
        parser = ArgumentParser("woven")
        parser.add_argument("-c", "--config", default = "config.json", help = "The path to the config file")
        args = parser.parse_args(namespace = WovenArgs)
        config_path = Path(args.config)
        config = fromdict(WovenConfig, loads(config_path.read_bytes()))
        generate_wg_configs(config)
    except ValueError as e:
        print(f"error: {e}", file = stderr)

if __name__ == "__main__":
    main()
